from pyramid.view import view_config
from pyramid.exceptions import HTTPNotFound
from .util import debug_log


DRS_VERSION_1 = 'v1'
DRS_PREFIX_V1 = f'ga4gh/drs/{DRS_VERSION_1}'
DRS_OBJECT_GET = DRS_PREFIX_V1 + '/objects/{object_id}'
DRS_OBJECT_GET_ACCESS_URL = DRS_PREFIX_V1 + '/objects/{object_id}/access/{access_id}'
DRS_OBJECT_GET_ACCESSS_URL_SLASH = DRS_PREFIX_V1 + '/objects/{object_id}/access/'
DRS_OBJECT_GET_ACCESSS_URL_NO_SLASH = DRS_PREFIX_V1 + '/objects/{object_id}/access'
REQUIRED_FIELDS = [
    'id',
    'created_time',
    'drs_id',
    'self_uri',
    'size',
    'checksums'
]
ACCESS_METHOD_REQUIRED_FIELDS = [
    'access_url',
    'type'
]


def includeme(config):
    config.add_route('drs_objects', '/' + DRS_OBJECT_GET)
    config.add_route('drs_download', '/' + DRS_OBJECT_GET_ACCESS_URL)
    config.add_route('drs_download_slash', '/' + DRS_OBJECT_GET_ACCESSS_URL_SLASH)
    config.add_route('drs_download_no_slash', '/' + DRS_OBJECT_GET_ACCESSS_URL_NO_SLASH)
    config.scan(__name__)


def validate_drs_object(drs_object):
    """ Validates the structure of a drs object (required fields)
        Because we're not wrapping in any object-oriented structure, the internal API
        will call this and throw a validation error if the returned DRS object
        does not conform to structure.
    """
    for required_key in REQUIRED_FIELDS:
        assert required_key in drs_object
    if 'access_methods' in drs_object:
        for required_key in ACCESS_METHOD_REQUIRED_FIELDS:
            for access_method in drs_object['access_methods']:
                assert required_key in access_method


def get_and_format_drs_object(request, object_uri):
    """ Call request.embed on the object_uri and reformats it such that it fits the DRS
        format, returning access_methods etc as needed if it is a file
    """
    try:
        drs_object = request.embed(object_uri, '@@drs', as_user=True)
    except Exception:
        raise HTTPNotFound('You accessed a DRS object_uri that either does not exist'
                           ' or you do not have access to it.')
    drs_object['self_uri'] = f'drs://{request.host}{request.path}'
    return drs_object


def get_drs_url(request, object_uri):
    """ Does 2 calls - one to verify the object_uri is in fact a valid DRS object and
        another to get the bytes of the DRS object """
    try:
        default_method = None
        requested_download_method = request.path.split('/')[-1]
        drs_obj = get_and_format_drs_object(request, object_uri)
        access_methods = drs_obj.get('access_methods', [])
        for access_method in access_methods:
            if access_method['type'] == 'https':  # prefer https
                default_method = access_method['access_url']
            if not default_method and access_method['type'] == 'http':  # allow http
                default_method = access_method['access_url']
            if access_method['type'] == requested_download_method:
                return access_method['access_url']
        else:
            if default_method:
                return default_method
            raise Exception  # no https default results in exception
    except Exception as e:
        raise HTTPNotFound(f'You accessed a DRS object that either you do not have access to,'
                           f' did not pass valid access_id or does not exist {str(e)}')


@view_config(
    route_name='drs_objects', request_method='GET'
)
@debug_log
def drs_objects(context, request):
    """ Implements DRS GET as specified by the API description
    https://ga4gh.github.io/data-repository-service-schemas/preview/release/drs-1.0.0/docs/#_getobject
    """
    drs_object_uri = '/' + request.matchdict['object_id']
    formatted_drs_object = get_and_format_drs_object(request, drs_object_uri)
    try:
        validate_drs_object(formatted_drs_object)
    except AssertionError as e:
        raise ValueError(f'Formatted DRS object does not conform to spec - check your @@drs'
                         f' implementation: {str(e)}')
    return formatted_drs_object


@view_config(route_name='drs_download_no_slash', request_method='GET')
@view_config(route_name='drs_download_slash', request_method='GET')
@view_config(route_name='drs_download', request_method='GET')
@debug_log
def drs_objects_download(context, request):
    """ Implements DRS GET bytes as specified by the API description
    https://ga4gh.github.io/data-repository-service-schemas/preview/release/drs-1.0.0/docs/#_getaccessurl

    NOTE: access_id is discarded - permissions are validated by @@download when navigated
    """
    drs_object_uri = '/' + request.matchdict['object_id']
    return get_drs_url(request, drs_object_uri)



import pytest
import urllib.parse

from dcicutils.lang_utils import n_of
from dcicutils.misc_utils import filtered_warnings
from dcicutils.qa_utils import MockResponse
from pyramid.testing import DummyRequest
from unittest import mock
from snovault import renderers
from snovault.renderers import (
    best_mime_type, should_transform, MIME_TYPES_SUPPORTED, MIME_TYPE_DEFAULT,
    MIME_TYPE_JSON, MIME_TYPE_HTML, MIME_TYPE_LD_JSON, MIME_TYPE_TRIAGE_MODE,
)


pytestmark = [pytest.mark.setone, pytest.mark.working]


class DummyResponse(MockResponse):

    def __init__(self, content_type=None, status_code: int = 200, json=None, content=None, url=None,
                 params=None):
        self.params = {} if params is None else params
        self.content_type = content_type
        super().__init__(status_code=status_code, json=json, content=content, url=url)


def test_mime_variables():

    # Really these don't need testing but it's useful visually to remind us of their values here.
    assert MIME_TYPE_HTML == 'text/html'
    assert MIME_TYPE_JSON == 'application/json'
    assert MIME_TYPE_LD_JSON == 'application/ld+json'

    # The MIME_TYPES_SUPPORTED is a list whose first element has elevated importance as we've structured things.
    # First check that it is a list, and that its contents contain the things we support. That isn't controversial.
    assert isinstance(MIME_TYPES_SUPPORTED, list)
    assert set(MIME_TYPES_SUPPORTED) == {MIME_TYPE_JSON, MIME_TYPE_HTML, MIME_TYPE_LD_JSON}
    # Check that the first element is consistent with the MIME_TYPE_DEFAULT.
    # It's an accident of history that this next relationship matters, but at this point check for consistency.
    assert MIME_TYPE_DEFAULT == MIME_TYPES_SUPPORTED[0]
    # Now we concern ourselves with the actual values...
    # TODO: I think it's a bug that JSON is at the head of this list (and so the default) in cgap-portal.
    #       cgap-portal needs to be made to match what Fourfront does to dig it out of a bug I introduced.
    #       -kmp 29-Jan-2022
    assert MIME_TYPES_SUPPORTED == [MIME_TYPE_JSON, MIME_TYPE_HTML, MIME_TYPE_LD_JSON]
    assert MIME_TYPE_DEFAULT == MIME_TYPE_JSON

    # Regardless of whether we're using legacy mode or modern mode, we should get the same result.
    assert MIME_TYPE_TRIAGE_MODE in ['legacy', 'modern']


VARIOUS_MIME_TYPES_TO_TEST = ['*/*', 'text/html', 'application/json', 'application/ld+json', 'text/xml', 'who/cares']


def test_best_mime_type():

    the_constant_answer = MIME_TYPE_DEFAULT

    with filtered_warnings("ignore", category=DeprecationWarning):
        # Suppresses this warning:
        #     DeprecationWarning: The behavior of .best_match for the Accept classes is currently being maintained
        #       for backward compatibility, but the method will be deprecated in the future, as its behavior is not
        #       specified in (and currently does not conform to) RFC 7231.

        for requested_mime_type in VARIOUS_MIME_TYPES_TO_TEST:
            req = DummyRequest(headers={'Accept': requested_mime_type})
            assert best_mime_type(req, 'legacy') == the_constant_answer
            assert best_mime_type(req, 'modern') == the_constant_answer
            req = DummyRequest(headers={})  # The Accept header in the request just isn't being consulted
            assert best_mime_type(req, 'modern') == the_constant_answer
            assert best_mime_type(req, 'modern') == the_constant_answer


TYPICAL_URLS = [
    'http://whatever/foo',
    'http://whatever/foo/',
    'http://whatever/foo.json',
    'http://whatever/foo.html',
]

ALLOWED_FRAMES_OR_NONE = ['raw', 'page', 'embedded', 'object', 'bad', None]

SOME_HTTP_METHODS = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'HEAD']

ALLOWED_FORMATS_OR_NONE = ['json', 'html', None]


def test_should_transform():

    passed = []
    failed = []
    problem_area = set()

    with filtered_warnings("ignore", category=DeprecationWarning):
        # Suppresses this warning:
        #     DeprecationWarning: The behavior of .best_match for the Accept classes is currently being maintained
        #       for backward compatibility, but the method will be deprecated in the future, as its behavior is not
        #       specified in (and currently does not conform to) RFC 7231.

        for method in SOME_HTTP_METHODS:
            for _format in ALLOWED_FORMATS_OR_NONE:
                for requested_mime_type in VARIOUS_MIME_TYPES_TO_TEST:
                    for response_content_type in VARIOUS_MIME_TYPES_TO_TEST:
                        for frame in [None] + ALLOWED_FRAMES_OR_NONE:
                            for url in TYPICAL_URLS:

                                params = {}
                                if frame is not None:
                                    params['frame'] = frame
                                if _format is not None:
                                    params['format'] = _format

                                req = DummyRequest(headers={'Accept': requested_mime_type},
                                                   method=method,
                                                   url=url,
                                                   params=params)
                                resp = DummyResponse(content_type=response_content_type, url=url)

                                _should_transform = should_transform(req, resp)

                                situation = {
                                    'method': method,
                                    "format": _format,
                                    "encoded_params": urllib.parse.urlencode(params),
                                    "requested": requested_mime_type,
                                    "response_content_type": response_content_type,
                                    "frame": frame,
                                    "url": url,
                                    "params": params,
                                }

                                if req.method not in ('GET', 'HEAD'):
                                    rule_applied = "method not GET or HEAD"
                                    correct = not _should_transform
                                elif resp.content_type != 'application/json':
                                    # If the response MIME type is not application/json,
                                    # it just can't be transformed at all.
                                    rule_applied = "content_type is not application/json"
                                    correct = not _should_transform
                                elif params.get("frame", "page") != 'page':
                                    rule_applied = "?frame=xxx is not page"
                                    correct = not _should_transform
                                elif _format is not None:
                                    rule_applied = "?format=xxx given but not html"
                                    correct = _should_transform is (_format == 'html')
                                else:
                                    # TODO: kmp thinks this behavior is a bug. It should default to HTML. -kmp 23-Mar-2021
                                    correct = _should_transform is False  # If no cue is given, default to JSON


                                if correct:
                                    # There are a lot of cases, so we don't print stuff here by default, but
                                    # uncomment to see what cases are passing as they pass:
                                    # print(situation)
                                    # print("=should_transform?=>", _should_transform, "(correct)")
                                    passed.append(situation)

                                else:
                                    # There are a lot of cases, so we don't print stuff here by default, but
                                    # uncomment to see what cases are failing as they fail:
                                    # print(situation)
                                    # print("=should_transform?=>", _should_transform, "(WRONG)")
                                    failed.append(situation)
                                    problem_area.add(rule_applied)

    if failed:
        # Collect all failures in one place:
        print("FAILED:")
        for failure in failed:
            print(" method=%(method)s format=%(format)s requested=%(requested)s"
                  " response_content_type=%(response_content_type)s frame=%(frame)s"
                  " url=%(url)s params=%(encoded_params)s"
                  % failure)

    n_failed = len(failed)
    n_passed = len(passed)
    assert not n_failed, (
            "%s passed, %s FAILED (%s: %s)"
            % (n_passed, n_failed, n_of(problem_area, "problem area"), ", ".join(problem_area))
    )
    print("\n", n_passed, "combinations tried. ALL PASSED")


def test_should_transform_without_best_mime_type():

    # As we call things now, we really don't need the best_mime_type function because it just returns the
    # first element of its first argument. That probably should change. Because it should be a function
    # of the request and its Accept offerings. Even so, we test for this now not because this makes programs
    # right, but so we notice if/when this truth changes. -kmp 23-Mar-2021

    with mock.patch.object(renderers, "best_mime_type") as mock_best_mime_type:

        # Demonstrate that best_mime_type(...) could be replaced by MIME_TYPES_SUPPORTED[0]
        mock_best_mime_type.return_value = MIME_TYPES_SUPPORTED[0]

        test_should_transform()



from dcicutils.misc_utils import get_setting_from_context

# Global snovault settings file, contained in the Settings class
# Declare default values for options below as part of the class, set them as needed in
# 'includeme' below.


def includeme(config):
    settings = config.registry.settings
    nested_is_set = get_setting_from_context(settings, 'mappings.use_nested', env_var='MAPPINGS_USE_NESTED')
    if nested_is_set:
        Settings.MAPPINGS_USE_NESTED = True


class Settings:
    MAPPINGS_USE_NESTED = False



from copy import deepcopy
from jsonschema import Draft202012Validator
from jsonschema import validators
from jsonschema.exceptions import ValidationError
from pyramid.settings import asbool
from pyramid.threadlocal import get_current_request
from pyramid.traversal import find_resource


NO_DEFAULT = object()


def get_resource_base(validator, linkTo):
    from snovault import COLLECTIONS
    request = get_current_request()
    collections = request.registry[COLLECTIONS]
    if validator.is_type(linkTo, 'string'):
        resource_base = collections.get(linkTo, request.root)
    else:
        resource_base = request.root
    return resource_base


def normalize_links(validator, links, linkTo):
    skip_links = (request := get_current_request()) and asbool(request.params.get('skip_links', False))
    resource_base = get_resource_base(validator, linkTo)
    normalized_links = []
    errors = []
    for link in links:
        try:
            # Difference between us and IGVF - it is possible in our system in a roundabout way to refer to a link
            # not by string but by dictionary, in most cases this shouldn't matter but does come up occasionally in
            # unit tests (looking for errors) - Will July 31 2023
            if isinstance(link, dict):
                for possible in ['uuid', '@id']:
                    if possible in link:
                        link = link[possible]
                        break
                else:
                    raise KeyError
            normalized_links.append(
                str(find_resource(resource_base, link.replace(':', '%3A')).uuid)
            )
        except KeyError:

            # 2024-02-13: To help out smaht-submitr refererential integrity checking,
            # include the schema type name (linkTo) as well as the idenitifying value (link).

            # 2024-02-21/dmichaels:
            # If skip_links then ignore reference/linkTo errors.
            # Currently ONLY used by smaht-submitr (via smaht-portal/loadxl_extensions.py).

            if not skip_links:
                errors.append(
                    ValidationError(f"Unable to resolve link: /{linkTo}/{link}")
                )
            normalized_links.append(
                link
            )
    return normalized_links, errors


def should_mutate_properties(validator, instance):
    if validator.is_type(instance, 'object'):
        return True
    return False


def get_items_or_empty_object(validator, subschema):
    items = subschema.get('items', {})
    if validator.is_type(items, 'object'):
        return items
    return {}


def maybe_normalize_links_to_uuids(validator, property, subschema, instance):
    errors = []
    if 'linkTo' in subschema:
        link = instance.get(property)
        if link:
            normalized_links, errors = normalize_links(
                validator,
                [link],
                subschema.get('linkTo'),
            )
            instance[property] = normalized_links[0]
    if 'linkTo' in get_items_or_empty_object(validator, subschema):
        links = instance.get(property, [])
        if links:
            normalized_links, errors = normalize_links(
                validator,
                links,
                subschema.get('items').get('linkTo'),
            )
            instance[property] = normalized_links
    for error in errors:
        yield error


def set_defaults(validator, property, subschema, instance):
    if 'default' in subschema:
        instance.setdefault(
            property,
            deepcopy(subschema['default'])
        )
    if 'serverDefault' in subschema:
        server_default = validator.server_default(
            instance,
            subschema
        )
        if server_default is not NO_DEFAULT:
            instance.setdefault(
                property,
                server_default
            )


def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS['properties']

    def mutate_properties(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if not validator.is_type(subschema, 'object'):
                continue
            yield from maybe_normalize_links_to_uuids(validator, property, subschema, instance)
            set_defaults(validator, property, subschema, instance)

    def before_properties_validation_hook(validator, properties, instance, schema):
        if should_mutate_properties(validator, instance):
            yield from mutate_properties(validator, properties, instance, schema)
        yield from validate_properties(validator, properties, instance, schema)

    return validators.extend(
        validator_class, {'properties': before_properties_validation_hook},
    )


ExtendedValidator = extend_with_default(Draft202012Validator)


class SerializingSchemaValidator(ExtendedValidator):

    SERVER_DEFAULTS = {}

    def add_server_defaults(self, server_defaults):
        self.SERVER_DEFAULTS.update(server_defaults)
        return self

    def serialize(self, instance):
        self._original_instance = instance
        self._mutated_instance = deepcopy(
            self._original_instance
        )
        errors = list(
            self.iter_errors(
                self._mutated_instance
            )
        )
        return self._mutated_instance, errors

    def server_default(self, instance, subschema):
        factory_name = subschema['serverDefault']
        factory = self.SERVER_DEFAULTS[factory_name]
        return factory(instance, subschema)



"""\
Examples
For the development.ini you must supply the paster app name:

    %(prog)s development.ini --app-name app --init --clear

"""

import argparse
import atexit
import logging
import os.path
import select
import shutil
import subprocess
import sys

from dcicutils.misc_utils import PRINT
from pyramid.paster import get_app, get_appsettings
from pyramid.path import DottedNameResolver
from .elasticsearch import create_mapping
from .project_app import app_project
from .tests import elasticsearch_fixture, postgresql_fixture


EPILOG = __doc__

logger = logging.getLogger(__name__)


def nginx_server_process(prefix='', echo=False):
    args = [
        os.path.join(prefix, 'nginx'),
        '-c', app_project().project_filename('nginx-dev.conf'),
        '-g', 'daemon off;'
    ]
    process = subprocess.Popen(
        args,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    if not echo:
        process.stdout.close()

    if echo:
        PRINT('Started: http://localhost:8000')

    return process


def ingestion_listener_compute_command(config_uri, app_name):
    return [
        'poetry', 'run', 'ingestion-listener', config_uri, '--app-name', app_name
    ]


def ingestion_listener_process(config_uri, app_name, echo=True):
    """ Uses Popen to start up the ingestion-listener. """
    args = ingestion_listener_compute_command(config_uri, app_name)

    process = subprocess.Popen(
        args,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    if echo:
        PRINT('Starting Ingestion Listener...')

    return process


def redis_server_process(echo=False):
    """ Handler that spins up a Redis server on port 6379 (default)"""
    args = [
        'redis-server',
        '--daemonize',
        'yes'
    ]
    process = subprocess.Popen(
        args,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    if not echo:
        process.stdout.close()
    if echo:
        print('Started Redis Server at redis://localhost:6379')
    return process


def main():
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description="Run development servers", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('config_uri', help="path to configfile")
    parser.add_argument('--clear', action="store_true", help="Clear existing data")
    parser.add_argument('--init', action="store_true", help="Init database")
    parser.add_argument('--load', action="store_true", help="Load test set")
    parser.add_argument('--datadir', default='/tmp/snovault', help="path to datadir")
    parser.add_argument('--no_ingest', action="store_true", default=False, help="Don't start the ingestion process.")
    args = parser.parse_args()

    run(app_name=args.app_name, config_uri=args.config_uri, datadir=args.datadir,
        # Ingestion is disabled. snovault has no such concept. -kmp 17-Feb-2023
        clear=args.clear, init=args.init, load=args.load, ingest=not args.no_ingest)


def run(app_name, config_uri, datadir, clear=False, init=False, load=False, ingest=True):

    #project = app_project(initialize=True)
    project = app_project()

    logging.basicConfig(format='')
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger(project.NAME).setLevel(logging.INFO)

    # get the config and see if we want to connect to non-local servers
    # TODO: This variable seems to not get used? -kmp 25-Jul-2020
    config = get_appsettings(config_uri, app_name)

    datadir = os.path.abspath(datadir)
    pgdata = os.path.join(datadir, 'pgdata')
    esdata = os.path.join(datadir, 'esdata')
    # ----- comment out from HERE...
    if clear:
        for dirname in [pgdata, esdata]:
            if os.path.exists(dirname):
                shutil.rmtree(dirname)
    if init:
        postgresql_fixture.initdb(pgdata, echo=True)
    # ----- ... to HERE to disable recreation of test db
    # ----- may have to `rm /tmp/snovault/pgdata/postmaster.pid`

    @atexit.register
    def cleanup_process():
        for process in processes:
            if process.poll() is None:
                process.terminate()
        for process in processes:
            try:
                for line in process.stdout:
                    sys.stdout.write(line.decode('utf-8'))
            except IOError:
                pass
            process.wait()

    processes = []

    # For now - required components
    postgres = postgresql_fixture.server_process(pgdata, echo=True)
    processes.append(postgres)

    es_server_url = config.get('elasticsearch.server', "localhost")

    if '127.0.0.1' in es_server_url or 'localhost' in es_server_url:
        # Bootup local ES server subprocess; otherwise assume we are connecting to a remote ES cluster.
        # The elasticsearch.server.actual_port property is useful (only) for running a localhost ElasticSearch
        # proxy in order to observe traffic (requests/responses) between portal and ElasticSearch with a tool like
        # mitmweb; e.g. setting elasticsearch.server.actual_port to 9201 and elasticsearch.server to localhost:9200
        # will cause ElasticSearch to actually run on port 9201 but will cause portal to talk to it via port 9200,
        # and then we can run mitmweb --mode reverse:http://localhost:9201 -p 9200 --web-port 8081 which will
        # allow us to browse to http://localhost:8081 locally to observe all of the ElasticSearch traffic.
        if (es_port := config.get('elasticsearch.server.actual_port', None)) and es_port.isdigit():
            es_port = int(es_port)
        elif ((colon := es_server_url.rfind(":")) > 0) and (es_port := es_server_url[colon + 1:]).isdigit():
            es_port = int(es_port)
        else:
            es_port = None
        elasticsearch = elasticsearch_fixture.server_process(esdata, port=es_port, echo=True)
        processes.append(elasticsearch)
    elif not config.get('indexer.namespace'):
        raise Exception(
            'It looks like are connecting to remote elasticsearch.server but no indexer.namespace is defined.')
    elif not config.get("elasticsearch.aws_auth", False):
        # TODO detect if connecting to AWS or not before raising an Exception.
        PRINT(
            'WARNING - elasticsearch.aws_auth is set to false.'
            ' Connection will fail if connecting to remote ES cluster on AWS.')

    nginx = nginx_server_process(echo=True)
    processes.append(nginx)

    app = get_app(config_uri, app_name)
    settings = app.registry.settings

    # Optional components
    if 'redis.server' in settings:
        redis = redis_server_process(echo=True)
        processes.append(redis)

    if ingest:
        ingestion_listener = ingestion_listener_process(config_uri, app_name)
        processes.append(ingestion_listener)

    # clear queues and initialize indices before loading data. No indexing yet.
    # this is needed for items with properties stored in ES
    if init:
        create_mapping.run(app, skip_indexing=True, purge_queue=False)

    if init and load:
        load_test_data = app.registry.settings.get('load_test_data')
        load_test_data = DottedNameResolver().resolve(load_test_data)
        load_res = load_test_data(app)
        if load_res:  # None if successful
            raise load_res

        # now clear the queues and queue items for indexing
        create_mapping.run(app, check_first=True, strict=True, purge_queue=False)

    PRINT('Started. ^C to exit.')

    stdouts = [p.stdout for p in processes]

    # Ugly should probably use threads instead
    while True:
        readable, writable, err = select.select(stdouts, [], stdouts, 5)
        for stdout in readable:
            for line in iter(stdout.readline, b''):
                sys.stdout.write(line.decode('utf-8'))
        if err:
            for stdout in err:
                for line in iter(stdout.readline, b''):
                    sys.stdout.write(line.decode('utf-8'))
            break


if __name__ == '__main__':
    main()



import logging
from copy import deepcopy
from posixpath import join
from pyramid.httpexceptions import HTTPNotFound, HTTPServerError
import pyramid.request
from .crud_views import collection_add as sno_collection_add
from .interfaces import COLLECTIONS, CONNECTION
from .pyramid_compat import (
    native_,
    unquote_bytes_to_wsgi,
)
from .resources import Collection
from .schema_utils import validate_request
from dcicutils.misc_utils import check_true

log = logging.getLogger(__name__)


def includeme(config):
    config.scan(__name__)
    config.add_renderer('null_renderer', NullRenderer)
    config.add_request_method(embed, 'embed')
    config.add_request_method(embed, 'invoke_view')
    config.add_request_method(lambda request: set(), '_linked_uuids', reify=True)
    config.add_request_method(lambda request: {}, '_sid_cache', reify=True)
    config.add_request_method(lambda request: {}, '_rev_linked_uuids_by_item', reify=True)
    config.add_request_method(lambda request: {}, '_aggregated_items', reify=True)
    config.add_request_method(lambda request: {}, '_aggregate_for', reify=True)
    config.add_request_method(lambda request: False, '_indexing_view', reify=True)
    config.add_request_method(lambda request: None, '__parent__', reify=True)


# really simple exception for when a primary indexed item gets HTTPNotFound
class MissingIndexItemException(Exception):
    pass


def make_subrequest(request, path, method='GET', json_body=None, inherit_user=False, inherit_registry=False):
    """
    Make a subrequest from a parent request given a request path.
    Copies request environ data for authentication. Handles making the path
    WSGI compatible. Optionally can take a JSON body to attach to subrequest.
    Used in _embed to form requests to invoke, and can also be used externally

    May be better to just pull out the resource through traversal and manually
    perform security checks.

    Args:
        request: current Request object
        path (str): path for the subrequest. Can include query string
        method (str): subrequest method, defaults to GET
        json_body (dict): optional dict to attach as json_body to subrequest
        inherit_user (bool) : optional to inherit remote_user from parent request
        inherit_registry (bool): optional to inherit registry from parent request

    Returns:
        Request: the subrequest
    """
    env = request.environ.copy()
    # handle path, include making wsgi compatible and splitting out query string
    path = unquote_bytes_to_wsgi(native_(path))
    if path and '?' in path:
        path_info, query_string = path.split('?', 1)
        path_info = path_info
    else:
        path_info = path
        query_string = ''
    env['PATH_INFO'] = path_info
    env['QUERY_STRING'] = query_string
    subreq = request.__class__(env, method=method, content_type=None)
    if json_body:
        subreq.json = json_body
    else:
        subreq.body = b''
    subreq.remove_conditional_headers()
    # XXX "This does not remove headers like If-Match"
    subreq.__parent__ = request
    if inherit_user:
        subreq.remote_user = request.remote_user
    if inherit_registry:
        subreq.registry = request.registry
    return subreq


def embed(request, *elements, **kw):
    """
    Incredibly important function that is central to getting views in snovault.
    Since it is a reified method on Request, you can call it like:
    `request.embed(<elements to be joined in path>)`
    This function handles propogation of important request attrs to subrequests,
    as well as caching of requests and grabbing attrs from the subreq result.

    Check connection.py and cache.py for details on the embed_cache

    NOTES:
        path is formed by joining all positional args
        as_user=True for current user
        Pass in fields_to_embed as a keyword arg

    Args:
        request: Request calling this method
        *elements: variable length positional args used to make path
        **kw: arbitrary keyword arguments

    Returns:
        result of the invoked request
    """
    # Should really be more careful about what gets included instead.
    # Cache cut response time from ~800ms to ~420ms.
    embed_cache = request.registry[CONNECTION].embed_cache
    as_user = kw.get('as_user')
    index_uuid = kw.get('index_uuid')
    path = join(*elements)
    # as_user controls whether or not the embed_cache is used
    # if request._indexing_view is True, always use the cache
    if as_user is not None and not request._indexing_view:
        cached = _embed(request, path, as_user)
    else:
        cached = embed_cache.get(path, None)
        if cached is None:
            # handle common cases of as_user, otherwise use what's given
            subreq_user = 'EMBED' if as_user is None else as_user
            cached = _embed(request, path, as_user=subreq_user)
            embed_cache[path] = cached

    # NOTE: if result was retrieved from ES, the following cached attrs will be
    # empty: _aggregated_items, _linked_uuids, _rev_linked_by_item
    result = deepcopy(cached['result'])

    # aggregated_items may be cached; if so, add them to the request
    # these conditions only fulfilled when using @@embedded and aggregated
    # items have NOT yet been processed (_aggregate_for is removed if so)
    if index_uuid and getattr(request, '_aggregate_for').get('uuid') == index_uuid:
        request._aggregated_items = cached['_aggregated_items']
        request._aggregate_for['uuid'] = None
    request._linked_uuids.update(cached['_linked_uuids'])
    request._sid_cache.update(cached['_sid_cache'])
    # this is required because rev_linked_uuids_by_item is formatted as
    # a dict keyed by item with value of set of uuids rev linking to that item
    for item, rev_links in cached['_rev_linked_by_item'].items():
        if item in request._rev_linked_uuids_by_item:
            request._rev_linked_uuids_by_item[item].update(rev_links)
        else:
            request._rev_linked_uuids_by_item[item] = rev_links
    return result


def _embed(request, path, as_user='EMBED'):
    """
    Helper function used in embed() that creates the subrequest and actually
    invokes it. Sets a number of attributes from the parent request and
    returns a dictionary containing the result and a number of attributes
    from the invoked subreq.

    Another consideration, now that we're purging items from the DB, is that
    primarily indexed items may be purged by the time they make it to _embed.
    Check if @@index-data is in the request path to ensure this is the case,
    and gracefully exit with MissingIndexItemException on HTTPNotFound if so.

    Args:
        request: Request object
        path (str): subrequest path to invoke
        as_user (str/bool): involved in setting subreq.remote_user

    Returns:
        dict containing the result and a number of subrequest attributes
    """
    # Carl: the subrequest is 'built' here, but not actually invoked
    subreq = make_subrequest(request, path)
    # these attributes are propogated across the subrequest
    subreq.override_renderer = 'null_renderer'
    subreq._indexing_view = request._indexing_view
    subreq._aggregate_for = request._aggregate_for
    subreq._aggregated_items = request._aggregated_items
    subreq._sid_cache = request._sid_cache
    if as_user is not True:
        if 'HTTP_COOKIE' in subreq.environ:
            del subreq.environ['HTTP_COOKIE']
        subreq.remote_user = as_user
    # _linked_uuids are populated in item_view_object of resource_views.py
    try:
        result = request.invoke_subrequest(subreq)
    except HTTPNotFound:
        if '@@index-data' in path:
            # the resource to index is missing; likely purged
            raise MissingIndexItemException(path)
        else:
            # the resource is unexpectedly missing
            raise KeyError(path)
    return {'result': result, '_linked_uuids': subreq._linked_uuids,
            '_rev_linked_by_item': subreq._rev_linked_uuids_by_item,
            '_aggregated_items': subreq._aggregated_items,
            '_sid_cache': subreq._sid_cache}


def subrequest_object(request, object_id):
    subreq = make_subrequest(request, "/" + object_id)
    subreq.headers['Accept'] = 'application/json'
    # Tweens are suppressed here because this is an internal call and doesn't need things like HTML processing.
    # -kmp 2-Feb-2021
    response = request.invoke_subrequest(subreq, use_tweens=False)
    if response.status_code >= 300:  # alas, the response from a pyramid subrequest has no .raise_for_status()
        raise HTTPServerError("Error obtaining object: %s" % object_id)
    object_json = response.json
    return object_json


def subrequest_item_creation(request: pyramid.request.Request, item_type: str, json_body: dict = None) -> dict:
    """
    Acting as proxy on behalf of request, this creates a new item of the given item_type with attributes per json_body.

    For example,

        subrequest_item_creation(request=request, item_type='NobelPrize',
                                 json_body={'category': 'peace', 'year': 2016))

    Args:
        request: the request on behalf of which this subrequest is done
        item_type: the name of the item item type to be created
        json_body: a python dictionary representing JSON containing data to use in initializing the newly created item

    Returns:
        a python dictionary (JSON description) of the item created

    """

    if json_body is None:
        json_body = {}
    collection_path = '/' + item_type
    method = 'POST'
    # json_utf8 = json.dumps(json_body).encode('utf-8')  # Unused, but here just in case
    check_true(not request.remote_user, "request.remote_user has %s before we set it." % request.remote_user)
    request.remote_user = 'EMBED'
    subrequest = make_subrequest(request=request, path=collection_path, method=method, json_body=json_body)
    subrequest.remote_user = 'EMBED'
    subrequest.registry = request.registry
    # Maybe...
    # validated = json_body.copy()
    # subrequest.validated = validated
    registry: Registry = subrequest.registry  # noQA - PyCharm can't tell subrequest.registry IS a Registry
    collection: Collection = registry[COLLECTIONS][item_type]
    check_true(subrequest.json_body, "subrequest.json_body is not properly initialized.")
    check_true(not subrequest.validated, "subrequest was unexpectedly validated already.")
    check_true(not subrequest.errors, "subrequest.errors already has errors before trying to validate.")
    check_true(subrequest.remote_user == request.remote_user,
               "Mismatch: subrequest.remote_user=%r request.remote_user=%r"
               % (subrequest.remote_user, request.remote_user))
    validate_request(schema=collection.type_info.schema, request=subrequest, data=json_body)
    if not subrequest.validated:
        return {
            "@type": ["Exception"],
            "errors": subrequest.errors
        }
    else:
        json_result: dict = sno_collection_add(context=collection, request=subrequest, render=False)
        return json_result


class NullRenderer:
    '''Sets result value directly as response.
    '''
    def __init__(self, info):
        pass

    def __call__(self, value, system):
        request = system.get('request')
        if request is None:
            return value
        request.response = value
        return None



import structlog
from .interfaces import REDIS
from dcicutils.redis_utils import RedisBase, create_redis_client


log = structlog.getLogger(__name__)


def includeme(config):
    registry = config.registry
    try:
        registry[REDIS] = RedisBase(create_redis_client(url=registry.settings['redis.server']))
    except Exception as e:
        log.error(f"Cannot create RedisBase object: {registry.settings.get('redis.server')}")
        log.error(str(e))
        registry[REDIS] = None


class RedisModel(object):
    """
    Model used by resources returned from RedisStorage
    Analogous to esstorage.CachedModel, storage.Resource

    Redis Keystore Structure:
        * 2 tiered namespace, first by environment then customizable within the environment.
        * 2 sub-embedded objects, each with a dirty bit
            * 'meta' holds item metadata along with the raw view (properties)
            * 'embedded' holds the latest embedded view of this item

        <env_namespace>:<namespace>:<uuid>:meta|embedded -> Redis HASH object

        NOTE: all objects are serialized into JSON

            meta : {
                "dirty": 0|1,
                "item_type": <type>,  # Sample for example,
                "properties":  { ... item properties ... }
                "propsheets": { ... item propsheets ... },
                "unique_keys": { ... item unique_keys ... },
                "links": { ... item links ... },
                "uuid": <uuid>,
                "sid": <sid>,
                "max_sid": <max_sid>
            }

            OR

            embedded : {
                "dirty" : 0|1,
                "view" : { ... view of item ... }
            }


    Note that this allows for future use in different <namespace>'s.

    NOT USED AT THIS TIME.
    """
    used_datastore = 'redis'

    def __init__(self, meta):
        """ Takes dictionary document 'meta' """
        self.meta = meta

    @property
    def item_type(self):  # TODO
        raise NotImplementedError

    @property
    def properties(self):  # TODO
        raise NotImplementedError

    @property
    def propsheets(self):  # TODO
        raise NotImplementedError

    @property
    def unique_keys(self):  # TODO
        raise NotImplementedError

    @property
    def links(self):  # TODO
        raise NotImplementedError

    @property
    def uuid(self):  # TODO
        """
        Return UUID object to be consistent with Resource.uuid
        """
        raise NotImplementedError

    @property
    def sid(self):  # TODO
        raise NotImplementedError

    @property
    def max_sid(self):  # TODO
        raise NotImplementedError


class RedisConnection(object):
    """ This class is meant to be the intermediary between the RedisBase class and the PickStorage class.
        Requires helper methods that take advantage of the above redis APIs.

        NOT USED AT THIS TIME.
    """

    def __init__(self, registry, default=None):
        self.registry = registry
        self.namespace = registry.settings.get('indexer.namespace', '')  # use indexer namespace as Redis 1st key
        self.redis = RedisBase(registry[REDIS])  # raw connection API
        self.__default = 'default' if default is None else default  # default key namespace, name is configurable

    def get_by_uuid(self, uuid: str, namespace=None):
        """ Looks up the given uuid in the Redis keystore. If no namespace is specified, the default namespace is used.

            :param uuid: uuid of the item to find
            :param namespace: key namespace to use
            :returns: the requested view
        """
        pass

    def get_by_unique_key(self, unique_key: str, name: str):
        """ Looks up the given item by unique_key : name

            :param unique_key: name of field
            :param name: value of field
            :returns: the requested view
        """
        pass

    def purge_uuid(self, rid, item_type):
        """ Purges the given uuid from Redis

            :param rid: resource id
            :param item_type: type of item
        """
        pass

    def __len__(self):
        """ Count # of documents in Redis. """
        pass

    def update(self, model, properties, sheets, unique_keys, links):
        """ Update method. """
        pass



APP_VERSION_REGISTRY_KEY = 'snovault.app_version'

# This order determines order that items will be mapped + added to the queue
# Can use item type (e.g. file_fastq) or class name (e.g. FileFastq)
# This order is not meaningful for snovault as it is not a standalone
# app. This value should be overridden in the downstream application
ITEM_INDEX_ORDER = []



import logging
import netaddr

from dcicutils.ff_utils import get_health_page
from dcicutils.log_utils import set_logging
from pyramid.config import Configurator
from pyramid.settings import asbool
from .local_roles import LocalRolesAuthorizationPolicy

from .app import app_version, session, configure_dbsession, changelogs, json_from_path
from .calculated import calculated_property  # noqa
from .config import abstract_collection, collection, root  # noqa
from .elasticsearch import APP_FACTORY
from .elasticsearch.interfaces import INVALIDATION_SCOPE_ENABLED
from .interfaces import *  # noqa
from .resources import AbstractCollection, Collection, Item, Resource, Root, display_title_schema  # noqa
from .schema_utils import load_schema  # noqa
from .upgrader import upgrade_step  # noqa


def includeme(config):
    config.include('pyramid_retry')
    config.include('pyramid_tm')
    config.include('snovault.authentication')
    config.include('snovault.util')
    config.include('snovault.drs')
    config.include('snovault.stats')
    config.include('snovault.batchupgrade')
    config.include('snovault.calculated')
    config.include('snovault.config')
    config.include('snovault.connection')
    config.include('snovault.custom_embed')
    config.include('snovault.embed')
    config.include('snovault.json_renderer')
    config.include('snovault.validation')
    config.include('snovault.predicates')
    config.include('snovault.invalidation')
    config.include('snovault.upgrader')
    config.include('snovault.aggregated_items')
    config.include('snovault.storage')
    config.include('snovault.typeinfo')
    config.include('snovault.types')
    config.include('snovault.resources')
    config.include('snovault.attachment')
    config.include('snovault.schema_graph')
    config.include('snovault.jsonld_context')
    config.include('snovault.schema_views')
    config.include('snovault.crud_views')
    config.include('snovault.indexing_views')
    config.include('snovault.resource_views')
    config.include('snovault.settings')
    config.include('snovault.server_defaults')
    config.include('snovault.routes')


def main(global_config, **local_config):
    """
    This function returns a Pyramid WSGI application.
    """
    settings = global_config
    settings.update(local_config)

    # adjust log levels for some annoying loggers
    lnames = ['boto', 'urllib', 'elasticsearch', 'dcicutils']
    for name in logging.Logger.manager.loggerDict:
        if any(logname in name for logname in lnames):
            logging.getLogger(name).setLevel(logging.WARNING)

    set_logging(in_prod=settings.get('production'))
    # set_logging(settings.get('elasticsearch.server'), settings.get('production'))

    # TODO - these need to be set for dummy app
    # settings['snovault.jsonld.namespaces'] = json_asset('snovault:schemas/namespaces.json')
    # settings['snovault.jsonld.terms_namespace'] = 'https://www.encodeproject.org/terms/'
    settings['snovault.jsonld.terms_prefix'] = 'snovault'

    config = Configurator(settings=settings)
    config.registry[APP_FACTORY] = main  # used by mp_indexer
    config.include(app_version)

    config.include('pyramid_multiauth')  # must be before calling set_authorization_policy
    # Override default authz policy set by pyramid_multiauth
    config.set_authorization_policy(LocalRolesAuthorizationPolicy())
    config.include(session)

    config.include(configure_dbsession)
    config.include('snovault')
    config.commit()  # commit so search can override listing

    config.include('.renderers')

    if settings.get('elasticsearch.server'):
        config.include('snovault.search.search')
        config.include('snovault.search.compound_search')

    # only include this stuff if we're testing
    if asbool(settings.get('testing', False)):
        config.include('snovault.tests.testing_views')
        config.include('snovault.tests.root')

        # in addition, enable invalidation scope for testing - but NOT by default
        settings[INVALIDATION_SCOPE_ENABLED] = True
    else:
        config.include('snovault.root')

    if 'elasticsearch.server' in config.registry.settings:
        config.include('snovault.elasticsearch')

    # configure redis server in production.ini
    if 'redis.server' in config.registry.settings:
        config.include('snovault.redis')

    config.include(changelogs)

    # TODO This is optional AWS only - possibly move to a plug-in
    aws_ip_ranges = json_from_path(settings.get('aws_ip_ranges_path'), {'prefixes': []})
    config.registry['aws_ipset'] = netaddr.IPSet(
        record['ip_prefix'] for record in aws_ip_ranges['prefixes'] if record['service'] == 'AMAZON')

    # cache mirror_health in registry if need be
    mirror = settings.get('mirror.env.name', None)
    if mirror:
        settings['mirror_health'] = get_health_page(ff_env=mirror)

    app = config.make_wsgi_app()

    return app



import random
import uuid

from dcicutils.misc_utils import exported, utc_now_str
from snovault.schema_validation import NO_DEFAULT
from pyramid.path import DottedNameResolver
from pyramid.threadlocal import get_current_request
from snovault.schema_utils import server_default
from .interfaces import COLLECTIONS  # , ROOT
from string import digits  # , ascii_uppercase
from .project_app import app_project
from .server_defaults_misc import add_last_modified, get_now
from .server_defaults_user import _userid, get_userid, get_user_resource

exported(
    COLLECTIONS,
    add_last_modified,
    get_now,
    get_userid,
    get_user_resource
)


ACCESSION_FACTORY = __name__ + ':accession_factory'
ACCESSION_TEST_PREFIX = 'TST'

# Only within snovault (only called from schema_formats.py) to get around
# app_project call at file scope (came up as circular import in smaht ingester).
# ACCESSION_PREFIX = app_project().ACCESSION_PREFIX
def GET_ACCESSION_PREFIX():
    return app_project().ACCESSION_PREFIX


def includeme(config):
    accession_factory = config.registry.settings.get('accession_factory')
    if accession_factory:
        factory = DottedNameResolver().resolve(accession_factory)
    else:
        factory = enc_accession
    config.registry[ACCESSION_FACTORY] = factory


# XXX: This stuff is all added based on the serverDefault identifier in the schemas
# removing it altogether will totally break our code


@server_default
def userid(instance, subschema):  # args required by jsonschema-serialize-fork
    return _userid()


@server_default
def now(instance, subschema):  # args required by jsonschema-serialize-fork
    return utc_now_str()


@server_default
def uuid4(instance, subschema):
    return str(uuid.uuid4())


@server_default
def accession(instance, subschema):
    if 'external_accession' in instance:
        return NO_DEFAULT
    request = get_current_request()
    factory = request.registry[ACCESSION_FACTORY]
    # With 17 576 000 options
    ATTEMPTS = 10
    for attempt in range(ATTEMPTS):
        new_accession = factory(subschema['accessionType'])
        if new_accession in request.root:
            continue
        return new_accession
    raise AssertionError("Free accession not found in %d attempts" % ATTEMPTS)


#FDN_ACCESSION_FORMAT = (digits, digits, digits, ascii_uppercase, ascii_uppercase, ascii_uppercase)
FDN_ACCESSION_FORMAT = ['ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789']*7

def enc_accession(accession_type):
    random_part = ''.join(random.choice(s) for s in FDN_ACCESSION_FORMAT)
    return GET_ACCESSION_PREFIX() + accession_type + random_part


TEST_ACCESSION_FORMAT = (digits, ) * 7


def test_accession(accession_type):
    """ Test accessions are generated on test.encodedcc.org
    """
    random_part = ''.join(random.choice(s) for s in TEST_ACCESSION_FORMAT)
    return 'TST' + accession_type + random_part
